searchData={"items":[{"type":"module","title":"Fine","doc":"Fine is a C++ library enabling more ergonomic NIFs, tailored to Elixir.\n\nErlang provides C API for implementing native functions\n([`erl_nif`](https://www.erlang.org/doc/apps/erts/erl_nif.html)).\nFine is not a replacement of the C API, instead it is designed as a\ncomplementary API, enhancing the developer experience when implementing\nNIFs in C++.","ref":"Fine.html"},{"type":"module","title":"Features - Fine","doc":"- Automatic encoding/decoding of NIF arguments and return value,\n  inferred from function signatures.\n\n- Smart pointer enabling safe management of resource objects.\n\n- Registering NIFs and resource types via simple annotations.\n\n- Support for encoding/decoding Elixir structs based on compile time\n  metadata.\n\n- Propagating C++ exceptions as Elixir exceptions, with support for\n  raising custom Elixir exceptions.\n\n- Creating all static atoms at load time.","ref":"Fine.html#module-features"},{"type":"module","title":"Motivation - Fine","doc":"Some projects make extensive use of NIFs, where using the C API results\nin a lot of boilerplate code and a set of ad-hoc helper functions that\nget copied from project to project. The main idea behind Fine is to\nreduce the friction of getting from Elixir to C++ and vice versa, so\nthat developers can focus on writing the actual native code.","ref":"Fine.html#module-motivation"},{"type":"module","title":"Requirements - Fine","doc":"Currently Fine requires C++17. The supported compilers include GCC,\nClang and MSVC.","ref":"Fine.html#module-requirements"},{"type":"module","title":"Installation - Fine","doc":"Add `Fine` as a dependency in your `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:fine, \"~> 0.1.0\", runtime: false}\n  ]\nend\n```\n\nModify your makefiles to look for Fine header files, similarly to the\nERTS ones. Also make sure to use at least C++17.\n\n```shell\n# GCC/Clang (Makefile)\nCPPFLAGS += -I$(FINE_INCLUDE_DIR)\nCPPFLAGS += -std=c++17\n\n# MSVC (Makefile.win)\nCPPFLAGS=$(CPPFLAGS) /I\"$(FINE_INCLUDE_DIR)\"\nCPPFLAGS=$(CPPFLAGS) /std:c++17\n```\n\nWhen using `elixir_make`, set `FINE_INCLUDE_DIR` like this:\n\n```elixir\ndef project do\n  [\n    ...,\n    make_env: fn -> %{\"FINE_INCLUDE_DIR\" => Fine.include_dir()} end\n  ]\nend\n```\n\nOtherwise, you can inline the dir to `deps/fine/include`.","ref":"Fine.html#module-installation"},{"type":"module","title":"Usage - Fine","doc":"A minimal NIF adding two numbers can be implemented like so:\n\n```c++\n#include  \n\nint64_t add(ErlNifEnv *env, int64_t x, int64_t y) {\n  return x + y;\n}\n\nFINE_NIF(add, 0);\n\nFINE_INIT(\"Elixir.MyLib.NIF\");\n```","ref":"Fine.html#module-usage"},{"type":"module","title":"Encoding/Decoding - Fine","doc":"Terms are automatically encoded and decoded at the NIF boundary based\non the function signature. In some cases, you may also want to invoke\nencode/decode directly:\n\n```c++\n// Encode\nauto message = std::string(\"hello world\");\nauto term = fine::encode(env, message);\n\n// Decode\nauto message = fine::decode (env, term);\n```\n\nFine provides implementations for the following types:\n\n| Type                                 | Encoder | Decoder |\n| ------------------------------------ | ------- | ------- |\n| `fine::Term`                         | x       | x       |\n| `int64_t`                            | x       | x       |\n| `uint64_t`                           | x       | x       |\n| `double`                             | x       | x       |\n| `bool`                               | x       | x       |\n| `ErlNifPid`                          | x       | x       |\n| `ErlNifBinary`                       | x       | x       |\n| `std::string`                        | x       | x       |\n| `fine::Atom`                         | x       | x       |\n| `std::nullopt_t`                     | x       |         |\n| `std::optional `                   | x       | x       |\n| `std::variant `              | x       | x       |\n| `std::tuple `                | x       | x       |\n| `std::vector `                     | x       | x       |\n| `std::map `                     | x       | x       |\n| `fine::ResourcePtr `               | x       | x       |\n| `T` with [struct metadata](#structs) | x       | x       |\n| `fine::Ok `                  | x       |         |\n| `fine::Error `               | x       |         |\n\n> #### ERL_NIF_TERM {: .warning}\n>\n> In some cases, you may want to define a NIF that accepts or returns\n> a term and effectively skip the encoding/decoding. However, the NIF\n> C API defines `ERL_NIF_TERM` as an alias for an integer type, which\n> may introduce an ambiguity for encoding/decoding. For this reason\n> Fine provides a wrapper type `fine::Term` and it should be used in\n> the NIF signature in those cases. `fine::Term` defines implicit\n> conversion to and from `ERL_NIF_TERM`, so it can be used with all\n> `enif_*` functions with no changes.\n\n> #### Binaries {: .info}\n>\n> `std::string` is just a sequence of `char`s and therefore it makes\n> for a good counterpart for Elixir binaries, regardless if we are\n> talking about UTF-8 encoded strings or arbitrary binaries.\n>\n> However, when dealing with large binaries, it is preferable for the\n> NIF to accept `ErlNifBinary` as arguments and deal with the raw data\n> explicitly, which is zero-copy. That said, keep in mind that `ErlNifBinary`\n> is read-only and only valid during the NIF call lifetime.\n>\n> Similarly, when returning large binaries, prefer creating the term\n> with `enif_make_new_binary` and returning `fine::Term`, as shown below.\n>\n> ```c++\n> fine::Term read_data(ErlNifEnv *env) {\n>   const char *buffer = ...;\n>   uint64_t size = ...;\n>\n>   ERL_NIF_TERM binary_term;\n>   auto binary_data = enif_make_new_binary(env, size, &binary_term);\n>   memcpy(binary_data, buffer, size);\n>\n>   return binary_term;\n> }\n> ```\n>\n> You can also return `ErlNifBinary` allocated with `enif_alloc_binary`,\n> but keep in mind that returning the binary converts it to term, which\n> in turn transfers the ownership, so you should not use that `ErlNifBinary`\n> after the NIF finishes.\n\nYou can extend encoding/decoding to work on custom types by defining\nthe following specializations:\n\n```c++\n// Note that the specialization must be defined in the `fine` namespace.\nnamespace fine {\n  template <> struct Decoder  {\n    static MyType decode(ErlNifEnv *env, const ERL_NIF_TERM &term) {\n      // ...\n    }\n  };\n\n  template <> struct Encoder  {\n    static ERL_NIF_TERM encode(ErlNifEnv *env, const MyType &value) {\n      // ...\n    }\n  };\n}\n```","ref":"Fine.html#module-encoding-decoding"},{"type":"module","title":"Resource objects - Fine","doc":"Resource objects is a mechanism for passing pointers to C++ data\nstructures to and from NIFs, and around your Elixir code. On the Elixir\nside those pointer surface as reference terms (`#Reference<...>`).\n\nFine provides a construction function `fine::make_resource (...)`,\nsimilar to `std::make_unique` and `std::make_shared` available in the\nC++ standard library. This function creates a new object of the type\n`T`, invoking its constructor with the given arguments and it returns\na smart pointer of type `fine::ResourcePtr `. The pointer is\nautomatically decoded and encoded as a reference term. It can also be\npassed around C++ code, automatically managing the reference count\n(similarly to `std::shared_ptr`).\n\nYou need to indicate that a given class can be used as a resource type\nvia the `FINE_RESOURCE` macro.\n\n```c++\n#include  \n\nclass Generator {\npublic:\n  Generator(uint64_t seed) { /* ... */ }\n  int64_t random_integer() { /* ... */ }\n  // ...\n};\n\nFINE_RESOURCE(Generator);\n\nfine::ResourcePtr  create_generator(ErlNifEnv *env, uint64_t seed) {\n  return fine::make_resource (seed);\n}\n\nFINE_NIF(create_generator, 0);\n\nint64_t random_integer(ErlNifEnv *env, fine::ResourcePtr  generator) {\n  return generator->random_integer();\n}\n\nFINE_NIF(random_integer, 0);\n\nFINE_INIT(\"Elixir.MyLib.NIF\");\n```\n\nOnce neither Elixir nor C++ holds a reference to the resource object,\nit gets destroyed. By default only the `T` type destructor is called.\nHowever, in some cases you may want to interact with NIF APIs as part\nof the destructor. In that case, you can implement a `destructor`\ncallback on `T`, which receives the relevant `ErlNifEnv`:\n\n```c++\nclass Generator {\n  // ...\n\n  void destructor(ErlNifEnv *env) {\n    // Example: send a message to some process using env\n  }\n};\n```\n\nIf defined, the `destructor` callback is called first, and then the\n`T` destructor is called as usual.\n\nOftentimes NIFs deal with classes from third-party packages, in which\ncase, you may not control how the objects are created and you cannot\nadd callbacks such as `destructor` to the implementation. If you run\ninto any of these limitations, you can define your own wrapper class,\nholding an object of the third-party class and implementing the desired\nconstruction/destruction on top.\n\nYou can use `fine::make_resource_binary(env, resource, data, size)`\nto create a binary term with memory managed by the resource.","ref":"Fine.html#module-resource-objects"},{"type":"module","title":"Structs - Fine","doc":"Elixir structs can be passed to and from NIFs. To do that, you need to\ndefine a corresponding C++ class that includes metadata fields used\nfor automatic encoding and decoding. The metadata consists of:\n\n- `module` - the Elixir struct name as an atom reference\n\n- `fields` - a mapping between Elixir struct and C++ class fields\n\n- `is_exception` (optional) - when defined as true, indicates the\n  Elixir struct is an exception\n\nFor example, given an Elixir struct `%MyLib.Point{x: integer, y: integer}`,\nyou could operate on it in the NIF, like this:\n\n```c++\n#include  \n\nnamespace atoms {\n  auto ElixirMyLibPoint = fine::Atom(\"Elixir.MyLib.Point\");\n  auto x = fine::Atom(\"x\");\n  auto y = fine::Atom(\"y\");\n}\n\nstruct ExPoint {\n  int64_t x;\n  int64_t y;\n\n  static constexpr auto module = &atoms::ElixirMyLibPoint;\n\n  static constexpr auto fields() {\n    return std::make_tuple(std::make_tuple(&ExPoint::x, &atoms::x),\n                           std::make_tuple(&ExPoint::y, &atoms::y));\n  }\n};\n\nExPoint point_reflection(ErlNifEnv *env, ExPoint point) {\n  return ExPoint{-point.x, -point.y};\n}\n\nFINE_NIF(point_reflection, 0);\n\nFINE_INIT(\"Elixir.MyLib.NIF\");\n```\n\nStructs can be particularly convenient when using NIF resource objects.\nWhen working with resources, it is common to have an Elixir struct\ncorresponding to the resource. In the previous `Generator` example,\nyou may define an Elixir struct such as `%MyLib.Generator{resource: reference}`.\nInstead of passing and returning the reference from the NIF, you can\npass and return the struct itself:\n\n```c++\n#include  \n\nclass Generator {\npublic:\n  Generator(uint64_t seed) { /* ... */ }\n  int64_t random_integer() { /* ... */ }\n  // ...\n};\n\nnamespace atoms {\n  auto ElixirMyLibGenerator = fine::Atom(\"Elixir.MyLib.Generator\");\n  auto resource = fine::Atom(\"resource\");\n}\n\nstruct ExGenerator {\n  fine::ResourcePtr  resource;\n\n  static constexpr auto module = &atoms::ElixirMyLibPoint;\n\n  static constexpr auto fields() {\n    return std::make_tuple(\n      std::make_tuple(&ExGenerator::resource, &atoms::resource),\n    );\n  }\n};\n\nExGenerator create_generator(ErlNifEnv *env, uint64_t seed) {\n  return ExGenerator{fine::make_resource (seed)};\n}\n\nFINE_NIF(create_generator, 0);\n\nint64_t random_integer(ErlNifEnv *env, ExGenerator ex_generator) {\n  return ex_generator.resource->random_integer();\n}\n\nFINE_NIF(random_integer, 0);\n\nFINE_INIT(\"Elixir.MyLib.NIF\");\n```","ref":"Fine.html#module-structs"},{"type":"module","title":"Exceptions - Fine","doc":"All C++ exceptions thrown within the NIF are caught and raised as\nElixir exceptions.\n\n```c++\nthrow std::runtime_error(\"something went wrong\");\n// ** (RuntimeError) something went wrong\n\nthrow std::invalid_argument(\"expected x, got y\");\n// ** (ArgumentError) expected x, got y\n\nthrow OtherError(...);\n// ** (RuntimeError) unknown exception thrown within NIF\n```\n\nAdditionally, you can use `fine::raise(env, value)` to raise exception,\nwhere `value` is encoded into a term and used as the exception. This\nis not particularly useful with regular types, however it can be used\nto raise custom Elixir exceptions. Consider the following exception:\n\n```elixir\ndefmodule MyLib.MyError do\n  defexception [:data]\n\n  @impl true\n  def message(error) do\n    \"got error with data #{error.data}\"\n  end\nend\n```\n\nFirst, we need to implement the corresponding C++ class:\n\n```c++\nnamespace atoms {\n  auto ElixirMyLibMyError = fine::Atom(\"Elixir.MyLib.MyError\");\n  auto data = fine::Atom(\"data\");\n}\n\nstruct ExMyError {\n  int64_t data;\n\n  static constexpr auto module = &atoms::ElixirMyLibMyError;\n\n  static constexpr auto fields() {\n    return std::make_tuple(\n        std::make_tuple(&ExMyError::data, &atoms::data));\n  }\n\n  static constexpr auto is_exception = true;\n};\n```\n\nThen, we can raise it anywhere in a NIF:\n\n```c++\nfine::raise(env, ExMyError{42})\n// ** (MyLib.MyError) got error with data 42\n```","ref":"Fine.html#module-exceptions"},{"type":"module","title":"Atoms - Fine","doc":"It is preferable to define atoms as static variables, this way the\ncorresponding terms are created once, at NIF load time.\n\n```c++\nnamespace atoms {\n  auto hello_world = fine::Atom(\"hello_world\");\n}\n```","ref":"Fine.html#module-atoms"},{"type":"module","title":"Result types - Fine","doc":"When it comes to NIFs, errors often indicate unexpected failures and\nraising an exception makes sense, however you may also want to handle\ncertain errors gracefully by returning `:ok`/`:error` tuples, similarly\nto usual Elixir functions. Fine provides `Ok ` and `Error `\ntypes for this purpose.\n\n```c++\nfine::Ok<>()\n// :ok\n\nfine::Ok (1)\n// {:ok, 1}\n\nfine::Error<>()\n// :error\n\nfine::Error (\"something went wrong\")\n// {:error, \"something went wrong\"}\n```\n\nYou can use `std::variant` to express a union of possible result types\na NIF may return:\n\n```c++\nstd::variant , fine::Error > find_meaning(ErlNifEnv *env) {\n  if (...) {\n    return fine::Error (\"something went wrong\");\n  }\n\n  return fine::Ok (42);\n}\n```\n\nNote that if you use a particular union frequently, it may be convenient\nto define a type alias with `using`/`typedef` to keep signatures brief.","ref":"Fine.html#module-result-types"},{"type":"function","title":"Fine.include_dir/0","doc":"Returns the directory with Fine header files.","ref":"Fine.html#include_dir/0"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.37.2"}}